 /* $Id: mapchan.c,v 1.17 2008/09/11 20:47:59 yuk Exp $ */

//#define SYSLOG
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <libgen.h>
#include <pty.h>
#include <utmp.h>
#ifdef SYSLOG
#include <syslog.h>
#endif

#define MAXBYTES 4
#ifdef UTF
#include <iconv.h>
#define ICONV_MAX 1024
#define utf_name "UTF8"

#ifndef TRUE
# define TRUE 1
#endif

#ifndef FALSE
# define FALSE 0
#endif

static iconv_t incode_table = NULL;
static iconv_t utf_table = NULL;
static unsigned int badchar = '*';
#endif

char *DEFAULT_MAP =	"trivial.map";
char *DEFAULT_PTY =	"-";
char *DEFAULT_SHELL =	"/bin/sh";
char *DEFAULT_LOGFILE = "mapchan.log";

#ifndef MAX_SHELL_ARGS
  #define MAX_SHELL_ARGS 16
#endif

#ifndef CPATH
  #define CPATH "/etc/mapchan"
#endif

#define BUFSIZE 32*1024
#define MAXPTYNAME 16
#define MAXPTYS 416

#define inputF 1
#define outputF 2
#define deadF  4
#define composeF 8 
#define bellF 16
#define controlF 32
#define escCF 64

#ifdef DEBUG
FILE *dbg;
#endif
 
// Global parameters
char *tbl_file, *log_file;
char *shell_prog;
char ttytable[MAXPTYS];
int autoprobe98 = 1;
int quiet = 0;
int debug = 0;
int pause_open = 0;

// Global variables
#define kw_num 7
char* kw[kw_num] = { "input", "output", "dead", "compose", "bell", "control", "esc(" } ;
unsigned char deadC = '\0';
unsigned char composeC = '\0';
unsigned char mapon = '\0';
unsigned char mapoff = '\0';
int was_dead = 0;
unsigned char **outtbl = NULL;
unsigned char **intbl = NULL;
unsigned char **deadtbl = NULL;
unsigned char **escCtbl = NULL; 
int clear_xonxoff = 1;

int tables_size = 0;
int fast_quit = 0;
int not_tty = 0;
int pty_fd;     /* descriptor master */
int tty_fd;     /* descriptor slave */
pid_t pid;	   /* child's PID      */

fd_set readfds; /* select flags     */


extern int optind, opterr, optopt;
extern char *optarg;

/*
         mapchan -q -s shell -f file -h -?
 */

#ifdef UTF
char usage[] = { "\nUsage:\n\nmapchan [options]\
\n\nwhere options are :\n\
\t-s shell	- shell, if not /bin/sh\n\
\t-f file	- input map-file\n\
\t-p /dev/ttyXX[-YY] | auto - terminal device name(s) or \"auto\" for autoprobing\n\
\t-o /dev/ttyXX	- allocate the terminal for output stream\n\
\t-x            - don't change XON/XOFF settings for pseudoterminal\n\
                  (by default - do as stty -ixon -ixoff) \n\
\t-u codeset    - convert the codeset into UTF-8\n\
\t-q		- force and quiet\n\
\t-P		- pause when open terminal\n\
\t-d [logfile]  - debug mode on. If logfile is omitted - use $TMPDIR/mapchan.log\n\
\t-h    - print this screen and exit.\n" };
#else
char usage[] = { "\nUsage:\n\nmapchan [options]\
\n\nwhere options are :\n\
\t-s shell	- shell, if not /bin/sh\n\
\t-f file	- input map-file\n\
\t-p /dev/ttyXX[-YY] | auto - terminal device name(s) or \"auto\" for autoprobing\n\
\t-o /dev/ttyXX	- allocate the terminal for output stream\n\
\t-x            - don't change XON/XOFF settings for pseudoterminal\n\
                  (by default - do as stty -ixon -ixoff) \n\
\t-q		- force and quiet\n\
\t-P		- pause when open terminal\n\
\t-d [logfile]  - debug mode on. If logfile is omitted - use $TMPDIR/mapchan.log\n\
\t-h    - print this screen and exit.\n" };
#endif

void help() {
  if (!quiet) printf("\r\nmapchan started with %d bytes in tables\r\n", tables_size);
}  

#ifdef NOBASENAME
char *basename(char *path) {
  char *p;
  if (p=strrchr(path, '/')) p++;
  else p=path;
  return(p);
}
#endif

struct bytecell  {
  struct bytecell *addr;
  unsigned int ch;
};

struct bytecell table0_utf2ch[256];
char table_ch2utf[128][MAXBYTES+1];

void init_0tables() {
  int l, cod;

  l = sizeof(char*) * 256;
  tables_size += l;
  if (outtbl == NULL) outtbl = calloc(256, sizeof(char*));
  for (cod=0; cod < 256; cod++) {
     table0_utf2ch[cod].ch = cod;
     table0_utf2ch[cod].addr = NULL;
  }
}

#ifdef UTF
int init_iconv(char *incode_name) {
  utf_table = iconv_open(utf_name, incode_name);
  if (utf_table == (iconv_t) -1)
      return 1;
  incode_table = iconv_open(incode_name, utf_name);
  if (incode_table == (iconv_t) -1)
      return 1;
  return 0;
}


/* this function create the chain of tables "utf -> 8bit char" */
int init_utf_tables(char *codeset) {
  int l, ch, cod, codx;
  char buf[1];
  char obuf[MAXBYTES];
  char *pi, *po;
  int ii, io;
  struct bytecell *current_table;

  if (codeset == NULL) return(0);
  
  if (init_iconv(codeset) !=0) {
    perror("Can't init iconv");
    return(1);
  }
  for (cod=0; cod < 128; cod++) {
    codx = cod | 0x80;
    l = (MAXBYTES+1) * sizeof(char);
    tables_size += l;
    outtbl[codx] = malloc(l);
    buf[0] = codx;
    ii = 1; io = MAXBYTES;
    pi = buf; po = obuf;
    if ((l = iconv(utf_table, &pi, &ii, &po, &io)) == -1) {
       /* no such char in UTF-8 */
       table0_utf2ch[codx].ch = badchar;
       table0_utf2ch[codx].addr = NULL;
       outtbl[codx][0] = '\0';
    } else {
      current_table = table0_utf2ch;
      io = MAXBYTES - io;
      while (ii<io) {
        ch = obuf[ii] & 0xff;
        outtbl[codx][ii++] = ch;
        if (ii != io) {
          /* continue the chain */
          if (current_table[ch].addr == NULL) {
             l = 256 * sizeof(struct bytecell);
             tables_size += l;
             current_table = current_table[ch].addr = malloc(l);
          } else 
             current_table = current_table[ch].addr;
        } else {
          /* last char in the chain */
          current_table[ch].addr = NULL;
          current_table[ch].ch = cod | 0x80;
        }
      }
      outtbl[codx][ii] = '\0';
    }
  }
  iconv_close(utf_table);
  iconv_close(incode_table);

  return(0);
}

#endif

int from_utf(char ch) {
  static struct bytecell *current_table = table0_utf2ch;
  int c;

//  c = ch & 0x7f;
  c = ch & 0xff;
  if (current_table[c].addr == NULL) {
    c = current_table[c].ch;
    current_table = table0_utf2ch;
  } else {
    current_table = current_table[c].addr;
    c = 0;
  }
  return(c);
}

/*
 * Screen deInitialization
 */
void
deinit_mscr()
{

#ifdef DEBUG
   fprintf(dbg, "Closing pty\n");
#endif   
   kill(pid, SIGHUP);
   close(tty_fd);
   close(pty_fd);
}

void
create_child()
{
   pid_t n;
   char *shell_args[MAX_SHELL_ARGS];
   int i = 0;
   
#ifdef EXEC_SHELL
   shell_args[i++] = "sh";
   if (strcmp(DEFAULT_SHELL, shell_prog) != 0) {
      shell_args[i++] = "-c";
      shell_args[i++] = shell_prog;
   }
   shell_args[i++] = NULL;
#else
   char *p;

   p=shell_prog; 
   while (*p) {
      while (isspace(*p)) if (*p=='\0') break;
                          else p++;
      if (*p!='\0') {
         shell_args[i++] = p;
         if (i > MAX_SHELL_ARGS - 1) {
            fprintf(stderr, "mapchan: too many args for shell\n");
            exit(1);
         }
         while (!isspace(*p)) if (*p=='\0') break;
                              else p++;
         if (*p!='\0') *(p++) = '\0';
      }
   }   
   shell_args[i] = NULL;                          
#endif

   n = fork();
   if (n < 0) {
       perror("fork");
       exit(1);
   }
   if (!n) {
       /* child */
       close(pty_fd);

       /* Set up the slave pty as a controlling terminal. */
       setsid();

       /* Create stdin, stdout, and stderr. */
       dup2(tty_fd, 0);
       dup2(tty_fd, 1);
       dup2(tty_fd, 2);
       if (tty_fd > 2)
               close(tty_fd);
               

       /* Now pass the file descriptors to the shell. */
#ifdef EXEC_SHELL
//printf("exec 2\n");
//for (i=0 ; shell_args[i] != NULL; i++) printf("%d) %s\n", i, shell_args[i]);

       if (execvp(DEFAULT_SHELL, shell_args) == -1) {
           perror(shell_prog);
           exit(1);
       }
#else
       if (execvp(shell_prog, shell_args) == -1) {
           perror(shell_prog);
           exit(1);
       }
#endif

/*
       if (execl(shell_prog, (char*)basename(shell_prog), NULL) == -1) {
           perror(shell_prog);
           exit(1);
       }
*/
   }

   /* parent */
   pid = n;
   close(tty_fd);

   FD_ZERO(&readfds);
   FD_SET(0, &readfds);
   FD_SET(pty_fd, &readfds);
}

int set_tty_size() {
  struct winsize ws;

  if (not_tty == 0) {
    if (ioctl(fileno(stdin), TIOCGWINSZ, (char *)&ws)) {
       perror("Get window size");
       return(1);
    }
    
    if (ioctl (pty_fd, TIOCSWINSZ, (char *)&ws)) {
       perror("Set window size");
       return(1);
    }
  }
  return(0);
}

void
sigresize(sig)
int sig;
{
  set_tty_size();
  signal(SIGWINCH, sigresize);
}

void
sigchld(sig)
int sig;
{
  pid_t pid;
  int status;
  
  pid = waitpid(WAIT_ANY, &status, WNOHANG);
#ifdef DEBUG  
  fprintf(dbg, "\nChild %d died with status %d \n", pid, status);
#endif
#ifdef SYSLOG
  syslog(LOG_INFO, "\nSIGCHLD: Child %d died with status %d \n", pid, status);
#endif
  fast_quit = 1;

  signal(SIGCHLD, sigchld);
}

void
sighup(sig)
{
  int i;

  fprintf(stderr, "mapchan is catching signal %d\n", sig);
#ifdef SYSLOG
  syslog(LOG_INFO, "Signal %d\n", sig);
#endif

  for (i=1; i<NSIG; i++) signal(i, SIG_DFL);
  fast_quit = 1;
  
}  

/*
 * Translation 
 *  from String in Esc-sequence definition language
 *  to character codes

   The language definitions :
   0777 - octal code
   0xAA - hexadecimal code
   99   - decimal code
   'A'  - character A
   A    - character A
 */

char *parse_cod      /* Return address of next word or NULL */
(char *ks,  /* Input string                  */
char *out)  /* here we return character code */
{
   int i;
   unsigned int c;
   
l_start:   
   i = ks[0];
   
   if ((i >= '1') && (i <= '9'))
      /* decimal number */
      *out=atoi(ks);
   else   
     switch (i) {
     case '0':
                 if (ks[1] == 'x') 
                    /* hexadecimal number */
                    sscanf(&ks[2], "%x", &c);
                 else
                    /* octal number */
                    sscanf(&ks[1], "%o", &c);
                 *out = c & 0xff;
                 break;
     case '\'':  if (ks[1] == '\\') {
                    ks = &ks[2];
                    goto l_start;
                 }
     		 else
     		    *out = ks[1];
                 break;
     case '#':   
                 *out = 0;
                 return(NULL);
                 break;
     default:    *out = i;
                 break;
     }
     
   i = 0;
   if (ks[0] != '\'') 
      while (!isspace(ks[i])) if (!ks[i]) return(NULL);
                              else i++;
   else {
      i++;
      while (ks[i] != '\'') if (!ks[i]) return(NULL);
                            else i++;
      i++;
   }
   while (isspace(ks[i])) if (!ks[i]) return(NULL);
                          else i++;
   
   if (!ks[i]) return(NULL);
   
   return(&ks[i]);
}

int
keyword(char *s) {
  int i;
  
  for (i=0; i<kw_num; i++) if (!strncmp(kw[i], s, strlen(kw[i]))) break;
  if (i == kw_num) i = 0;
  else i = 1 << i;
  return(i);
}

/*
 * Decode table loading
 *
 */
int load_map(fn)
char *fn;       /* Input file name  */
{
  register int i, j, l;
  char bufin[BUFSIZE], buf[BUFSIZE];
  int flag;
  FILE *mf;
  char *p;
  unsigned char c;
  
  if (!fn) {
    for (i=0; i<128; i++) {
      p = malloc(2);
      p[0] = i;
      p[1] = '\0';
      outtbl[i] = (unsigned char*) p;
    }
    tables_size += 128*2;
    return(0);
  }

  flag = 0;
  
  for(i=0;isgraph(fn[i]); i++);
  fn[i] = '\0';
  if ((mf=fopen(fn, "r")) == NULL) {       /* open input file */
     /* no in current directory - let's find in CPATH */
     strncpy(buf, CPATH, BUFSIZE-2);
     strcat(buf, "/");
     strncat(buf, fn, BUFSIZE-strlen(buf)-2);
     if ((mf=fopen(buf, "r")) == NULL) {
        /* and no here too - error */
        perror(buf);
        return(1);
     }
  }
  /* Read definition string  */
  j = 0;
  while (fgets(bufin, 100, mf) != NULL) {
     j++;
     if (isspace(bufin[0]) || (bufin[0] == '#') || (!strlen(bufin)))
        continue;
     i=keyword(bufin);
     if (i != 0)
        switch (i) {
        case inputF:
                    	flag = i;
                    	intbl = (unsigned char**) -1;
                    	break;
        case outputF:
                    	flag = i;
                    	break;
        case escCF:
                    	flag = i;
                    	l = sizeof(struct bytecell) * 256;
                    	tables_size += l;
                    	escCtbl = malloc(l);
                    	break;
        case controlF:  
        		flag = i;
        		break;
        case bellF:
        case composeF:
                        break;
        case deadF:
        		if (flag & inputF) flag |= i;
        		else {
          		   fprintf(stderr, "error in file %s, str. %d\n",
        		           fn, j);
        		   return(1);
        		}
        		for (i=0; i<strlen(bufin); i++) 
        		  if (p=parse_cod(bufin, &c))
        		     parse_cod(p, &deadC);
                    	l = sizeof(struct bytecell) * 256;
                    	tables_size += l;
                        table0_utf2ch[deadC].addr = malloc(l);
                        
        		break;
        }
     else {
        p = parse_cod(bufin, &c);
        for (i=0; p != NULL; i++) {
          if (i>=99) {
             buf[i] = '\0';
             break;
          }
          p=parse_cod(p, &buf[i]);
        }
        if (flag == outputF) {
           outtbl[c] = malloc(i);
           tables_size += i;
           memcpy(outtbl[c], buf, i);
           outtbl[c][i] = '\0';
        }
        else if (flag & inputF) {
                if (i != 1) fprintf(stderr, "Warning: multibyte sequences in 'input' section are not supported\n");
                if (flag & deadF) {
                   table0_utf2ch[deadC].ch = buf[0];
                   table0_utf2ch[deadC].addr = NULL;
                }
                else {
                   table0_utf2ch[c].ch = buf[0];
                   table0_utf2ch[c].addr = NULL;
                }
             }
        else if (flag == controlF) {
           switch (c) {
           case '-':	/* mapping off */
                mapoff = buf[0];
                break;
           case '+':	/* mapping on */
                mapon = buf[0];
                break;
           default:
           	break;
           }
        }
        else if (flag == escCF) {
           escCtbl[c] = malloc(i);
           tables_size += i;
           memcpy(escCtbl[c], buf, i);
           escCtbl[c][i] = '\0';
        }
     }        
  }      
  fclose(mf);
  return(0);
}        

/*
 * Transforming of pseudo-terminal index into symbolic name:
 *
 * 0 -> /dev/ttya0, ... 415 -> /dev/ttyzf
 *
 */

char *idx2pty(int idx) {
  static char ptyname[MAXPTYNAME];
  char c1, c2;
  //               0123456789
  strcpy(ptyname, "/dev/ttya0");
  c1 = (idx >> 4) & 0xFF;
  if (c1 > 25) return(NULL);
  c2 = idx & 0xF;
  if (c2 > 9) c2 += 'a' - '9' - 1;
  ptyname[8] += c1;
  ptyname[9] += c2;
  return(ptyname);
}

/*
 * Getting pseudo-terminal index from its symbolic name
 *
 * /dev/ttya0 -> 0 ... /dev/ttyzf -> 415
 *
 */

int pty2idx(char *pty) {
  int c1, c2;
  int i;
  i=strlen(pty)-1;
  c2 = (pty[i--] - '0');
  if (c2 > 9) c2 -= 'a' - '9' - 1;
  if (c2 > 0xF) return(-1);
  c1 = (pty[i] - 'a') &0xFF;
  if (c1 > 25) return(-1);
  c1 <<= 4;
  return(c1+c2);
}


int parse_tty_name(char *tty) {
  int i, j;
  char buf[MAXPTYNAME];
  int i1, i2;
  int eol=0;
  char *p;

  i1 = 0;
  while (!eol) {
     i2 = i1;
     while (tty[i2] != ',') {
       if (tty[i2] == '\0') {
          eol++;
          break;
       }
       i2++;
     }
     tty[i2] = '\0';
     p = &tty[i1];
     if ((strncmp(p, "-", 1) == 0) || (strncmp(p, "unix98", 6) == 0)) { // autoprobe in Unix'98 format
       autoprobe98 = 1;
       break;
     }
     else {
       autoprobe98 = 0;
       if (strncmp(p, "auto", 4) == 0) { // autoprobe in traditional format
          for (i=0; i<MAXPTYS; i++) ttytable[i] = 1;
          break;
       }
       else if (strncmp(p, "/dev/tty", 8) == 0) {
          if (strlen(p) == 10) {   // exactly 1 pty name
             if ((i = pty2idx(p)) == -1) return(1);
             else ttytable[i] = 1;
          }
          else if (p[10] == '-') { // range of pty names
             sprintf(buf, "%.10s", p);
             i = pty2idx(buf);
             j = pty2idx(&p[11]);
             if ((i == -1) || (j == -1) || (j < i)) return(1);
             for (;i <= j;i++) ttytable[i] = 1;
          }
       }
     }

     i1 = i2 + 1;
  }
  
  
  return(0);
}

/* 
Select field number 'field_num' (starting from 1) from string str. 
Field - alphanumeric sequence what limited by spaces or EOL
If field_num < 0 - then limited only by EOL
Field as string with length not more then 'maxlen'
is returned into array 'field'

If no the field with such number or it is "-", 
then into 'field' returned default value 'fdefault' if it's not NULL.

If field is '+' - the function exit code is 1, else - 0.

*/
int get_field(char *str, char *field, char *fdefault, int field_num, int maxlen) {
  int i=0;
  int si=0;
  int c;
  int toend = 0;
#define skip_spaces() while ( ((c=str[si]) != '\0') && (isspace(c)) ) si++;
#define skip_text()   while ( ((c=str[si]) != '\0') && (!isspace(c)) ) si++;
  
  if (field_num < 0) {
     field_num = - field_num;
     toend++;
  }
  skip_spaces();
  i = 1;
  while (i != field_num) {
     skip_text();
     skip_spaces();
     i++;
  }
  i = 0;
  while ((c=str[si]) != '\0') {
     if (!toend && isspace(c) ) break;
     if (i < maxlen) 
        field[i++] = c;
     si++;
  }
  field[i] = '\0';
  if ( ((strlen(field) == 0) || ((strlen(field) == 1) && field[0] == '-'))
       && (fdefault != NULL) ) strncpy(field, fdefault, maxlen);
  
  if ((strlen(field) == 1) && (field[0] == '+')) i = 1;
  else i = 0;
    
  return(i);
}

void default_tbl() {
/*   1		   2		   3		4	5	6	   7
# Terminal         Map	           PTY	  	Quiet   Debug	LogFile    Shell
/dev/tty1	dos2koi.map	/dev/ttyc1	  -	  -	   -	   /bin/sh
*/
  char buf[BUFSIZE], buf1[BUFSIZE];
  char *p;
  FILE *f;
  
  if (p=(char*)ttyname(fileno(stdin))) {
     p=(char*)basename(p);
     strncpy(buf, CPATH, BUFSIZE-1);
     strncat(buf, "/default", BUFSIZE-strlen(buf));
     if (f=fopen(buf, "r")) {
        while (fgets(buf, BUFSIZE, f)) {
          /* Read file /etc/mapchan/default */
          
          // field 1 - control terminal name
          get_field(buf, buf1, NULL, 1, BUFSIZE);
          if (strcmp(p, buf1) == 0) {
             // found the current terminal name
             // field 2 - map-file name
             get_field(buf, buf1, DEFAULT_MAP, 2, BUFSIZE);
             tbl_file = malloc(strlen(buf1)+1);
             strcpy(tbl_file, buf1);
             // field 3 - pseudo-tty name
             get_field(buf, buf1, DEFAULT_PTY, 3, MAXPTYNAME);
             parse_tty_name(buf1);
             // field 4 - "quiet" mode
             quiet = get_field(buf, buf1, NULL, 4, 1); 
             // field 5 - debug mode
             debug = get_field(buf, buf1, NULL, 5, 1); 
             // field 6 - log file
             get_field(buf, buf1, DEFAULT_LOGFILE, 6, BUFSIZE);
             if (debug) {
                log_file = malloc(strlen(buf1) + 1);
                strcpy(log_file, buf1);
             }
             // field 7 - shell
             get_field(buf, buf1, DEFAULT_SHELL, -7, BUFSIZE); 
             shell_prog = malloc(strlen(buf1)+1);
             strcpy(shell_prog, buf1);
             break;
          }
        }
        fclose(f);
     }
  }
  else not_tty = 1;
}  
          
#define flush_in() while (j > 0) { \
                                m = write(pty_fd, p, j); \
                                p += m; j -= m; \
                              }

#define flush_out() while (j > 0) { \
                                   m = write(1, p, j); \
                                   p += m; j -= m; \
                                  }

char *
pty_name(char *tty_name)
{
        int l;
        char c1, c2;  //   0123456789
        static char n[] = { "/dev/ptyXX" };
        
        l = strlen(tty_name);
        c1 = tty_name[l-2];
        c2 = tty_name[l-1];
        n[8] = c1;
        n[9] = c2;
        return(n);
}

int main(argc, argv)
int argc;
char **argv;

{
        int i, m, n, j, k;
        char buf[BUFSIZE], buf1[BUFSIZE];
        unsigned char c;
        char *p, *p1;
        fd_set lreadfds; /* select flags     */
        struct termios tios, old_tios;
        int errflg;
        int domapping = 1;
        int out_mapping = 0;
        int in_mapping = 0;
        int status = 0;
        int tty_i;
        char *tty_name_i;
        int was_esc= 0;
        int was_esc_C = 0;
        char *output_tty = NULL;
#ifdef  LINUX_GRAPHICS_COMPAT
        int escm = 0;
        int escm_i = 0;
        char escm_arg[80];
#endif
#ifdef UTF
        int do_utf = FALSE;
        char *utf_incode = NULL;
#endif

#ifdef SYSLOG
    openlog("mapchan", LOG_PID, LOG_USER);
    syslog(LOG_INFO, "Started.");
#endif

#ifdef DEBUG
        if ((dbg = fopen("/dev/tty", "w")) == NULL) {
           perror("debug open /dev/tty");
           exit(1);
        }
#endif


        /*
         *  Command string parsing
         */

        /* Defaults */
        if ((p = getenv("TMPDIR")) == NULL) p = "/tmp";
        i = strlen(p) + strlen(DEFAULT_LOGFILE) + 1;
        p1 = malloc(i);
        sprintf(p1, "%s%s", p, DEFAULT_LOGFILE);
        DEFAULT_LOGFILE = log_file = p1;
        
        shell_prog = DEFAULT_SHELL;
        tbl_file = NULL;
        
        default_tbl();

        /* Keys scanning    */
        errflg = 0;
        while ((i = getopt(argc, argv, "qPxd::p:A:Z:s:f:o:u:h?")) != EOF) {
          switch (i) {

          case 'p': /* Explicit terminal name */
                    parse_tty_name(optarg);
                    break;
          case 's': /* what shell */
                    shell_prog = optarg;
                    break;
          case 'f': /* map-table */
                    tbl_file = optarg;
                    break;
          case 'q':
                    quiet = 1;
                    break;
          case 'x':
                    clear_xonxoff = 0;
                    break;
          case 'P':
                    pause_open = 1;
                    break;
          case 'd':
                    debug = 1;
                    if (optarg != NULL) log_file = optarg;
                    break;
          case 'o': if (optarg != NULL) output_tty = optarg;
                    break;
#ifdef UTF
          case 'u': do_utf = TRUE;
                    utf_incode = optarg;
                    break;
#endif
          case 'h':
          case '?':
                    errflg++;
                    break;
                    
          default:  
                    fprintf(stderr, "Unknown key: -%c\n", i);
                    errflg++;
                    break;
          }
        }
        if (errflg) {                 /* If there were errors 		*/
           fprintf(stderr, usage);    /*   show help text             	*/
           exit(1);                   /*   and exit with error status 	*/
        }
        
        if (output_tty) {
           if (freopen(output_tty, "r+", stdout) == NULL) {
              perror(output_tty);
              exit(1);
           }
           if (freopen(output_tty, "r+", stderr) == NULL) {
              perror(output_tty);
              exit(1);
           }
        }
        
       init_0tables();

#ifdef UTF
        if (init_utf_tables(utf_incode) != 0) {
             fprintf(stderr, "Can't create UTF chain table for codeset %s\n", utf_incode);
             exit(1);
        }
#endif
	if (load_map(tbl_file) ||
	   ((outtbl == NULL) && (intbl == NULL) && (deadtbl == NULL))) {
	   if (quiet) {
#ifdef EXEC_SHELL
//printf("exec 1\n");
              if (execlp("/bin/sh", "sh", "-c", shell_prog, NULL) == -1) {
#else

              if (execl(shell_prog, (char*)basename(shell_prog), NULL) == -1) {
#endif
                  perror(shell_prog);
                  exit(1);
              }
           } else {
 	      fprintf(stderr, "You don't define no one table - mapchan not needed\n");
              fprintf(stderr, usage);    /* Show help text            */
	      exit(1);
           }
	}   

#ifdef DEBUG
 if (outtbl) {
  printf("output\n");
  for (i=0; i<256; i++) if (outtbl[i]) { 
                           printf("   %c=%#0x) ", i, i);
                           for (j=0; j<strlen(outtbl[i]); j++)
                             if (outtbl[i][j] <= '\040') 
                                printf("%o ", outtbl[i][j] & 0377);
                             else
                                printf("%c ", outtbl[i][j]);
                           putchar('\n');
                         }
 } 
 if (deadtbl) {
  printf("dead\n");
  for (i=0; i<256; i++) if (deadtbl[i]) { 
                           printf("   %c=%#0x) ", i, i);
                           for (j=0; j<strlen(deadtbl[i]); j++)
                             if (deadtbl[i][j] <= '\040') 
                                printf("%o ", deadtbl[i][j] & 0377);
                             else
                                printf("%c ", deadtbl[i][j]);
                           putchar('\n');
                         }
 } 
#endif        
        if (not_tty == 0) {
          tcgetattr(0, &old_tios);
          memcpy(&tios, &old_tios, sizeof(struct termios));
  //        tios.c_iflag = 
        if (clear_xonxoff)
          tios.c_iflag &= ~(ICRNL | IXON | IXOFF);
        else
          tios.c_iflag &= ~(ICRNL);
          tios.c_oflag = tios.c_lflag = 0;
  //        tios.c_cflag |= CS8 ;
  //        tios.c_cflag &= ~( PARENB | PARODD | CSTOPB );
            
          tcsetattr(0, TCSANOW, &tios);
          if (pause_open) {
             sleep(1);
             putchar(' ');
             sleep(1);
          }
        }
        signal(SIGCHLD, sigchld);
        signal(SIGWINCH, sigresize);
        for (i=1; i<NSIG; i++) if ((i != SIGCHLD) && (i != SIGWINCH)) signal(i, sighup);
        
        if ( autoprobe98 ) { /* autoprobing in format Unix98 */
	   if (openpty(&pty_fd,&tty_fd,NULL,NULL,NULL)) {
		 perror("openpty failed: "); 
		 goto quit ; 
	   }
        } else  { /* exactly terminal name or traditional autoprobing */
             tty_i=0;
             while (tty_i < MAXPTYS) {
               if (ttytable[tty_i]) {
                 tty_name_i = idx2pty(tty_i);
#ifdef DEBUG
                 fprintf(dbg, "\rTrying %s\n\r", tty_name_i);
#endif
                 pty_fd = open(pty_name(tty_name_i), O_RDWR);
                 if (pty_fd >= 0) {
                    i++;
                    break;
                 }
               }
               tty_i++;
             }
             if (tty_i == MAXPTYS) {
               fprintf(stderr, "%s ERROR: No free ptys.\n\r", argv[0]);
               status = 1;
               goto quit;
             }
             
             tty_fd = open(tty_name_i, O_RDWR);
             if (tty_fd < 0) {
                perror(tty_name_i);
                status = 1;
                goto quit;
             }
        }
        
        if (not_tty == 0) {
           tcsetattr(tty_fd, TCSANOW, &old_tios);
           set_tty_size();
        }
        /* create screens */
        create_child();
        
        i = 0;
        help();
        
        out_mapping = outtbl || escCtbl ;
#ifdef UTF
        out_mapping = out_mapping || do_utf;
#endif
        in_mapping = intbl || deadtbl;
#ifdef UTF
        in_mapping = in_mapping || do_utf;
#endif
        
        while (!fast_quit) {
                lreadfds = readfds;
                j = 0;
                while (select(32, &lreadfds, (fd_set *) 0, 
                          (fd_set *) 0, (struct timeval *) 0) < 0) {
                   sleep(1);
                   if (j++ > 10) {
                     perror("select");
                     break;
                   }
                }
                if (j>10) {
                 break;
                }
                if (FD_ISSET(0, &lreadfds)) {
                        n = read(0, buf, sizeof buf);
#ifdef SYSLOG
                        syslog(LOG_INFO, "read0 %d", n);
#endif
                        if (n <= 0) {
                                perror("read 0");
                                break;
                        }
                        else if (in_mapping) {
                               p = buf1;
                               for (i=0,j=0;i<n;i++) {
                                 c = buf[i];
                                 if ((c = from_utf(c)) == 0) continue;
                                 if (j+1 >= BUFSIZE)
                                       flush_in();
                                  buf1[j++] = c;
                               }
                             }  
                             else { 
                               j = n;
                               p = buf;
                             }
                             flush_in();
                }

                if (FD_ISSET(pty_fd, &lreadfds)) {
                        n = read(pty_fd, buf, sizeof buf);
#ifdef SYSLOG
                        syslog(LOG_INFO, "read1 %d", n);
#endif
                        if (!n) break;
                        if (n < 0) {
                                if (errno != EIO)
                                        perror("read pty_fd");
                                continue;/*break;*/
                        }
                        else if (out_mapping) {
                               p = buf1;
                               for (i=0, j=0;i<n;i++) {
                                 c = buf[i];
#ifdef MAP_ONOFF
                                 if (domapping) {
                                   if (mapoff && (c == mapoff)) {
                                      domapping = 0;
                                      continue;
                                   }
                                 }
                                 else {
                                   if (mapon && (c == mapon)) {
                                      domapping = 1;
                                      continue;
                                   }
                                 }
#endif                                 

#ifdef ESC_C_TBL
                                 if (domapping && (escCtbl != NULL)) {
                                    if (was_esc) {
                                       was_esc = 0;
                                       if (c == '(') {
                                          was_esc_C++;
                                          continue;
                                       }
                                       else {
                                         if (j >= BUFSIZE-1) flush_out();
                                         buf1[j++] = '\033';
                                       }
                                    }
                                    if (was_esc_C) {
                                       was_esc_C = 0;
                                       if (escCtbl[c]) {
                                          if (j+strlen(escCtbl[c]) >= BUFSIZE) flush_out();
                                          k = 0;
                                          while (escCtbl[c][k]) buf1[j++] = escCtbl[c][k++];
                                          continue;
                                       }
                                       else {
                                          if (j >= BUFSIZE-2) flush_out();
                                          buf1[j++] = '\033';
                                          buf1[j++] = '(';
                                       }
                                    }
                                 
                                    if (c == '\033') {
                                       was_esc++;
                                       continue;
                                    }
                                 }
#endif

                                 if (domapping && ((p1=outtbl[c]) != NULL)) {
                                      if (j+strlen(p1) >= BUFSIZE) {
                                         flush_out();
                                      }
                                      k = 0;
                                      while (p1[k]) buf1[j++] = p1[k++];
                                 }
                                 else {
                                   if (j+1 >= BUFSIZE) {
                                      flush_out();
                                   }
                                   buf1[j++] = buf[i];
                                 }
#ifdef  LINUX_GRAPHICS_COMPAT
                                 /*  
                                     linux console pseudographics compatibility:
                                     \033 [ 11 m - turn off mapping
                                     \033 [ 12 m - turn off mapping
                                     \033 [ 10 m - turn on mapping
                                     TODO: control array overflow!!!
                                 */
                                 if (escm == 0) {
                                    if (c == '\033') escm++;
                                 } else if (escm == 1) {
                                    if (c == '[') { escm++; escm_i = 0; }
                                    else escm = 0;
                                 } else if (escm >= 2) {
                                    if (isdigit(c)) escm_arg[escm_i++] = c;
                                    else {
                                         escm_arg[escm_i] = '\0';
                                         escm_i = 0;
                                         if ((c == ';') || (c == 'm')) {
                                                   switch (atoi(escm_arg)) {
                                                   case 10: escm = 3;
                                                            break;
                                                   case 11: 
                                                   case 12: 
                                                            escm = 4;
                                                            break;
                                                   default: 
                                                            break;
                                                   }
                                                   if ((c == 'm') &&  (escm > 2)) {
                                                      domapping = escm & 1;
                                                      escm = 0;
                                                   }
                                                   escm_i = 0;
                                                   escm_arg[0] = '\0';
                                         }
                                         else escm = 0;
                                    }
                                 }
#endif
                               }
                             }
                             else {
                               p = buf;
                               j = n;
                             }
                             flush_out();
                }
        }

        deinit_mscr();
quit:
        if (not_tty == 0)
          if (tcsetattr(0, TCSANOW, &old_tios) < 0)
                perror("tcsetattr oldtios");

#ifdef SYSLOG
        syslog(LOG_INFO, "Ended.");
        closelog();
#endif
        if (!quiet) printf("\r\nmapchan ended.\r\n");
        return(status);
}

